import java.util.Random;
/*
 * This program implements a genetic algorithm to find good members.  
 * The members consist of genes of the alphabet of digits (0..9).   
 */

public class genetic {
	
	private static final int POPULATION_SIZE = 20; //needs to be even
	private static final int PARENTS_BREEDING = 5;
	private static final int GENES_IN_MEMBER = 10;
	private static final double MUTATION_RATE = 0.05; 
	
	private static final int GENERATIONS_TO_RUN = 40; //80;
	
	private static int[] searchValue = {5,1,5,8,6,7,5,3,0,9};
	
	//private static final int WORST_VALUE = (9*GENES_IN_MEMBER)+1;
	private static final int WORST_VALUE = -1;
	
	private static Random randomNumberGenerator = new Random(3);
	
	//This is the basic data of the program.
	private static int[][] population = new int[POPULATION_SIZE][GENES_IN_MEMBER];
	private static int[][] tempPopulation = new int[POPULATION_SIZE][GENES_IN_MEMBER];
	private static int tempPopulationOffset = 0;
	private static int[] bestParents = new int [PARENTS_BREEDING];
	

	//Put the initial values into the population.
	//Side-Effect: population is set.
	private static void initializePopulation () {
		for (int memberNumber = 0; memberNumber < POPULATION_SIZE; memberNumber++) {
			for (int geneNumber = 0; geneNumber < GENES_IN_MEMBER; geneNumber++) {
				population[memberNumber][geneNumber] = randomNumberGenerator.nextInt(10);
			}
		}
	}
	
	//Print out the entire population.
	private static void printPopulation (int[][]population, boolean includeEvaluation) {
		System.out.println("*****");
		for (int memberNumber = 0; memberNumber < POPULATION_SIZE; memberNumber++) {
			for (int geneNumber = 0; geneNumber < GENES_IN_MEMBER; geneNumber++) {
				System.out.print(population[memberNumber][geneNumber] +" ");
			}
			if (includeEvaluation ) {
				int value = evaluateMember(memberNumber);
				System.out.print(value);
			}
			System.out.println();
		}
	}

	//Select whether to return absolute sum or distance from the searchValue;
	private static int evaluateMember(int memberOffset) {
		//return (evaluateMemberDistanceFromSearch(memberOffset));
		return (evaluateMemberSum(memberOffset));
	}

	//Calculate the sum of all the digits 
	private static int evaluateMemberSum(int memberOffset) {
		int result = 0;
		for (int geneNumber = 0; geneNumber < GENES_IN_MEMBER; geneNumber++) {
			result += population[memberOffset][geneNumber];
		}
		return result;
	}
	
	//Find the distance from a member to the searchValue.
	private static int evaluateMemberDistanceFromSearch(int memberOffset) {
		int result = 0;
		for (int geneNumber = 0; geneNumber < GENES_IN_MEMBER; geneNumber++) {
			int difference = population[memberOffset][geneNumber]-searchValue[geneNumber];
			if (difference < 0) difference = difference * -1;  //get the absolute value
			result += difference;
		}
		return result;
	}

	//Use the tempPopulation and add two new children from two parents.
	private static void crossOver(int mom, int dad) {
		int crossOverPoint = randomNumberGenerator.nextInt(GENES_IN_MEMBER-2);
		crossOverPoint += 1;
		//System.out.println(mom + " " + dad + " " + crossOverPoint);
		for (int geneNumber = 0; geneNumber < GENES_IN_MEMBER; geneNumber++) {
			if (geneNumber < crossOverPoint) {
				tempPopulation[tempPopulationOffset][geneNumber] = population[mom][geneNumber];
				tempPopulation[tempPopulationOffset+1][geneNumber] = population[dad][geneNumber];
			}
			else {
				tempPopulation[tempPopulationOffset][geneNumber] = population[dad][geneNumber];
				tempPopulation[tempPopulationOffset+1][geneNumber] = population[mom][geneNumber];
			}	
		}
		tempPopulationOffset += 2;
	}
	
	//Just take the temporary population, and make it the new population.
	private static void copyTempToPopulation() {
		for (int memberNumber = 0; memberNumber < POPULATION_SIZE; memberNumber++) {
			for (int geneNumber = 0; geneNumber < GENES_IN_MEMBER; geneNumber++) {
				population[memberNumber][geneNumber] = tempPopulation[memberNumber][geneNumber];
			}
		}
		tempPopulationOffset = 0;
	}
	
	//Print out the indices of the best parents.
	private static void printBestParents() {
		System.out.println("Best Parents");
		for (int offset = 0; offset < PARENTS_BREEDING; offset ++) {
			System.out.print(bestParents[offset] + " ");
		}
		System.out.println();
	}
	
	//**********Functions for determining if one gene is better than another
	//Use this if you want to maximise the string.  You also need to set WORST_VALUE (small)
	//firstIsBetterLarge
	private static boolean firstIsBetter (int first, int second) {
		if (first >= second)
			return true;
		else
			return false;
	}
	
	//Use this if you want to minimise the string.  You also need to set WORST_VALUE (large)
	//firstIsBetterSmall
	private static boolean firstIsBetterSmall (int first, int second) {
		if (first < second)
			return true;
		else
			return false;
	}
	//*******
	
	//Just set the best values to values worse than possible.  It's determined by the 
	//constant WORST_VALUE.
	private static int[] initBestValues () {
		int[] returnValue  = new int [PARENTS_BREEDING];
		for (int bestOffset = 0; bestOffset < PARENTS_BREEDING; bestOffset ++) {
			returnValue[bestOffset] = WORST_VALUE;
		}
		
		return (returnValue);
	}
	
	//Get the best PATENTS_BREEDING parents based on the firstIsBetter function.
	//Side-Effect
	//This is stored in the class variable bestParents.
	private static void setBestParents() {
		int [] bestValues = initBestValues();
		
		for (int memberNumber = 0; memberNumber < POPULATION_SIZE; memberNumber++) {
			int thisMemberValue = evaluateMember(memberNumber);
			int place = -1;
			
			//Find out if and where this value is among the best.
			//It may not be in this list in which case place remains -1.
			for (int bestOffset = 0; bestOffset < PARENTS_BREEDING; bestOffset ++) {
				if ((place < 0) && (firstIsBetter(thisMemberValue,bestValues[bestOffset])))
					place = bestOffset;
			}
			
			//Insert this and move down
			if (place >=0)  {
				for (int bestOffset = PARENTS_BREEDING-1; bestOffset > place; bestOffset --) {
					bestParents[bestOffset] = bestParents[bestOffset-1];
					bestValues[bestOffset]=bestValues[bestOffset-1];
				}
				bestParents[place] = memberNumber;
				bestValues[place]=thisMemberValue;
			}
		}
		printBestParents();
	}
	
	
	//Randomly select two of the best parents.  May have the same mom and dad.
	private static int [] getParents () {
		int [] parents = new int[2];
		int mom = randomNumberGenerator.nextInt(PARENTS_BREEDING);
		mom=bestParents[mom];
		int dad = randomNumberGenerator.nextInt(PARENTS_BREEDING);
		dad=bestParents[dad];
		
		parents[0] = mom;
		parents[1] = dad;
		
		return parents;
	}
	
	//Each gene in each member is randomly changed with a low rate.  The rate is determined
	//by the constant mutate.
	//Side-Effect, population is changed.
	private static void mutate () {
		for (int memberNumber = 0; memberNumber < POPULATION_SIZE; memberNumber++) {
			for (int geneNumber = 0; geneNumber < GENES_IN_MEMBER; geneNumber++) {
				double mutationRoll = randomNumberGenerator.nextDouble();
				if (mutationRoll < MUTATION_RATE) {
					int newNumber = randomNumberGenerator.nextInt(10);
					population[memberNumber][geneNumber] = newNumber;
					//System.out.println("mutate " + memberNumber + " " + geneNumber + " " + newNumber);
				}
			}
		}
	}
	
	//Use cross over and mutation to get a new generation from the old one.
	private static void getNextGeneration() {
		
		//Sexual Reproduction
		setBestParents();
		for (int parentPair = 0; parentPair < POPULATION_SIZE/2; parentPair++) {
			int [] parents = getParents();
			crossOver(parents[0],parents[1]);
		}
		copyTempToPopulation();
		
		//Mutation 
		mutate();
		
	}
	
	
	//This is the entry point to the program.  
	//It loops through generations making (hopefully) better and better members.
	public static void main(String[] args) {
		System.out.println("hello world");
		initializePopulation();
		printPopulation(population,true);

		for (int generations = 0; generations < GENERATIONS_TO_RUN; generations++) {
			getNextGeneration();
			printPopulation(population,true);
		}

		
		System.out.println("goodbye cruel world");
	}

}
